Attribute VB_Name = "modRandString"
' MODULE:       modRandString
' FUNCTION:     Contains functions for generation of random strings given regular expression templates


' For simplicity's sake, parentheses, brackets, and braces will be referred to as parentheses
'   until we actually do actions specific to them


' SYNTAX OF REGULAR EXPRESSIONS USED IN THIS MODULE
' SYMBOL        MEANING     EXAMPLE
' "|"           Or          "a|b" returns either "a" or "b"
' "(",")"       Separator   "(a|b)(c|d)" returns either "ac", "ad", "bc", or "bd"
' "[","]"       Optional    "[a|b] has a 50% chance of returning nothing.
'                           Otherwise, it is equivalent to "(a|b)"
' "{","}"       Replace     If "{A}" is set to "ab," then "{A}c" returns "abc"


' FUNCTION:     RegExpRandomString
' PURPOSE:      Generates a random string given a regular expression template
' PARAMETERS:   Template - String, template using regular expressions
'               Replacements - 2-column array, holds replacement strings
' RETURNS:      String, random string given regular expression templates
Public Function RegExpRandomString(Template As String, Optional Replacements As Variant) As String
    Dim CurPos As Long              ' Holds current position in string parsing
    Dim CurChar As String           ' Holds current character in string parsing
    Dim ParenLevel As Long          ' Open parentheses:  Increment by 1, Close parentheses:  Decrease by 1
    Dim StartParenPos As Long       ' Position of left outermost starting parentheses
    Dim EndParenPos As Long         ' Position of left outermost end parentheses
    Dim ParenContents As String     ' Contents of left outermost parentheses
    Dim ParseArray As Variant       ' Holds results of parsing
    Dim i As Long                   ' General-purpose looping variable
    
    ' Before we proceed, let's check the syntax of the template string to avoid errors later on
    If Not CheckSyntax(Template) Then
        ' I personally dislike raising errors, so just return "ERROR", and then Debug.Print
        
        RegExpRandomString = "ERROR"
        Debug.Print "Syntax error"
        Exit Function
    End If
    
    Randomize
    
    ' I HATE how UBound handles uninitialized variant arrays, so let's initialize it
    '   before VBstarts fussing about it
    ReDim ParseArray(0)
    
    ' Keep executing the loop until there are no more parentheses
    Do Until (InStr(1, Template, "(") + InStr(1, Template, "[") + InStr(1, Template, "{")) = 0
        
        ' Start by looping through the string, finding the left outer-most parentheses contents
        For CurPos = 1 To Len(Template)
            CurChar = Mid(Template, CurPos, 1)
            
            ' Encountered an opening parentheses/bracket/braces, so increment ParenLevel
            ' We want to retrieve the string contents between an opening parentheses
            '   with a ParenLevel of 1 and a closing parentheses that decrements
            '   ParenLevel to zero
            If InStr(1, "([{", CurChar) Then
                ParenLevel = ParenLevel + 1
                If ParenLevel = 1 Then StartParenPos = CurPos
            End If
            
            ' Encountered a closing parentheses/bracket/braces, so decrement ParenLevel
            If InStr(1, ")]}", CurChar) Then
                ParenLevel = ParenLevel - 1
                
                ' If ParenLevel decrements to zero, that means we found the closing parentheses
                '   that corresponds with StartParenPos. We can exit the loop here.
                If ParenLevel = 0 Then
                    EndParenPos = CurPos
                    Exit For
                End If
                
            End If
            
        Next CurPos
        
        ' Now obtain the contents of the left outermost parentheses
        ParenContents = Mid(Template, StartParenPos, EndParenPos - StartParenPos + 1)
        'Debug.Print ParenContents
        
        ' Custom handling depending on whether it is a parentheses, brackets, or braces
        Select Case Left(ParenContents, 1)
        Case "("
            
            ' For parentheses, simply strip the parentheses and then recurse
            ParenContents = Mid(ParenContents, 2, Len(ParenContents) - 2)
            ParenContents = RegExpRandomString(ParenContents, Replacements)
            
        Case "["
            
            ' Half the time, the stuff inside the brackets will not be parsed
            If (Rnd > 0.5) Then
                ParenContents = ""
            Else
                ParenContents = Mid(ParenContents, 2, Len(ParenContents) - 2)
                ParenContents = RegExpRandomString(ParenContents, Replacements)
            End If
            
        Case "{"
        
            ' Strip the braces and expand them
            ParenContents = Mid(ParenContents, 2, Len(ParenContents) - 2)
            ParenContents = RegExpRandomString(ParenContents, Replacements)
            Select Case UCase(ParenContents)
            Case "C"
                ' Built-in replacement for consonants
                
                ParenContents = "(b|c|d|f|g|h|j|k|l|m|n|p|q|r|s|t|v|w|x|y|z)"
            Case "V"
                ' Built-in replacement for vowels
                
                ParenContents = "(a|e|i|o|u)"
            Case "L"
                ' Built-in replacement for all letters
                
                ParenContents = "(a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z)"
            Case "#"
                ' Built-in replacement for single-digit numbers
                
                ParenContents = "(1|2|3|4|5|6|7|8|9|0)"
            Case Else
                ' The Replacements parameter holds a 2-column array.
                '   The first column holds the string sought (i.e. "V")
                '   while the second holds the value that replaces it (i.e. "(a|e|i|o|u)").
                '   Make ABSOLUTELY sure that the replacement is within parentheses, brackets,
                '   or braces (i.e. "(a|b)" or nasty things will happen! To simplify things,
                '   "(((((a|b)))))" is the same as "(a|b)"
                
                For i = 0 To UBound(Replacements, 2)
                    If ParenContents = Replacements(0, i) Then ParenContents = Replacements(1, i)
                Next i
            End Select
        End Select
        
        ' The final step is to replace the contents with the new
        Template = Left(Template, StartParenPos - 1) & ParenContents & Right(Template, Len(Template) - EndParenPos)
    
    Loop
    
    'Now, after the loop is done, template has been modified to be free of parentheses,
    '  so now we can finally parse the "|"'s which are left over.
    ParseArray = Split(Template, "|")
    If UBound(ParseArray) > -1 Then
        RegExpRandomString = ParseArray(Int(Rnd * (UBound(ParseArray) + 1)))
    Else
        RegExpRandomString = Template
    End If
    
End Function

' FUNCTION:     CheckSyntax
' PURPOSE:      Checks to see if a template's syntax is correct. Does NOT check for
'               placement of parentheses, brackets, or braces (i.e. CheckSyntax will
'               return true for "([a)]"). However, if used in RegExpRandomString,
'               the errror will be detected because CheckSyntax is run at every stage
'               of parsing.
' PARAMETERS:   Template - String, text to be analyzed
' RETURNS:      Boolean, true if correct, false otherwise

Public Function CheckSyntax(Template As String) As Boolean
    Dim OpeningParen As Long        ' Number of opening parentheses
    Dim ClosingParen As Long        ' Number of closing parentheses
    Dim OpeningBrack As Long        ' Number of opening brackets
    Dim ClosingBrack As Long        ' Number of closing brackets
    Dim OpeningBrace As Long        ' Number of opening braces
    Dim ClosingBrace As Long        ' Number of closing braces
    Dim i As Long                   ' General-purpose looping variable
    
    For i = 1 To Len(Template)
        Select Case Mid(Template, i, 1)
        Case "("
            OpeningParen = OpeningParen + 1
        Case ")"
            ClosingParen = ClosingParen + 1
        Case "["
            OpeningBrack = OpeningBrack + 1
        Case "]"
            ClosingBrack = ClosingBrack + 1
        Case "{"
            OpeningBrace = OpeningBrace + 1
        Case "}"
            ClosingBrace = ClosingBrace + 1
        End Select
    Next i
    
    If (OpeningParen = ClosingParen) And (OpeningBrack = ClosingBrack) And (OpeningBrace = ClosingBrace) Then
        CheckSyntax = True
    Else
        CheckSyntax = False
    End If
End Function
